import json

import click
from http_types import HttpExchangeBuilder

from meeshkan.schemabuilder.result import BuildResult
from meeshkan.convert.pcap import convert_pcap

from .schemabuilder import build_schema_online
from .schemabuilder.writer import write_build_result
from .config import setup
from .logger import get as getLogger


LOGGER = getLogger(__name__)


def log(*args):
    LOGGER.debug(*args)


@click.group()
def cli():
    """
    Meeshkan CLI.
    """
    setup()  # Ensure setup is done before invoking the CLI.


@click.command()
@click.option("-i", "--input-file", required=True, type=click.File('rb'), help="Input file. Use dash '-' to read from stdin.")
@click.option("-o", "--out", required=False, default='out', type=click.Path(exists=False, file_okay=False, writable=True, readable=True), help="Output directory. If the directory does not exist, it is created if the parent directory exists.")
def build(input_file, out):
    """
    Build OpenAPI schema from recordings.
    """
    requests = (HttpExchangeBuilder.from_dict(
        json.loads(line)) for line in input_file)

    schema = build_schema_online(requests)

    build_result = BuildResult(openapi=schema)

    log("Result: %s", json.dumps(schema))
    write_build_result(out, build_result)


@click.command()
@click.option("-i", "--input-file", required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=True), help="Path to a packet capture file.")
@click.option("-o", "--out", required=False, default='recordings.jsonl', type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, allow_dash=True), help="Output file.")
def convert(input_file, out):
    """
    Convert recordings from PCAP to JSONL format.
    """

    if not input_file.endswith('.pcap'):
        raise ValueError(
            'Only .pcap files are accepted as input. Got: {}'.format(input_file))

    request_response_pairs = convert_pcap(input_file)

    log("Writing to: %s", out)

    counter = 0
    with open(out, 'w') as f:
        for reqres in request_response_pairs:
            f.write((json.dumps(reqres) + "\n"))
            counter += 1

    log("Wrote %d lines.", counter)


cli.add_command(build)  # type: ignore
cli.add_command(convert)  # type: ignore

if __name__ == '__main__':
    cli()



import copy
from genson import SchemaBuilder  # type: ignore


def _to_openapi_compatible(schema):
    """OpenAPI does not strictly use JSON schema so convert to the OpenAPI subset.

    https://swagger.io/docs/specification/data-models/keywords/

    Arguments:
        schema -- Schema

    Returns:
        Modified schema
    """
    schema = copy.deepcopy(schema)
    del schema['$schema']
    return schema


def to_json_schema(obj, schema=None):
    """Create JSON schema based on an object.

    Arguments:
        obj {dict} -- Dictionary object

    Keyword Arguments:
        schema {dict} -- Existing schema if exists. (default: {None})

    Returns:
        [dict] -- New or updated schema.
    """
    builder = SchemaBuilder()
    if schema is not None:
        builder.add_schema(schema)
    builder.add_object(obj)
    schema = builder.to_schema()
    return schema


def to_openapi_json_schema(obj, schema=None):
    """Create OpenAPI-compatible JSON schema from object.

    https://swagger.io/docs/specification/data-models/keywords/

    Arguments:
        obj {[type]} -- [description]

    Keyword Arguments:
        schema {[type]} -- [description] (default: {None})

    Returns:
        [type] -- [description]
    """
    schema = to_json_schema(obj, schema)
    openapi_compatible_schema = _to_openapi_compatible(schema)
    return openapi_compatible_schema



import copy
from http_types import HttpExchange as HttpExchange
from ..logger import get as getLogger
from functools import reduce
from typing import Any, List, Iterator, cast, Tuple, Optional
from openapi_typed import Info, MediaType, OpenAPIObject, PathItem, Response, Operation, Schema
from typeguard import check_type  # type: ignore
import json
from typing_extensions import Literal
from .json_schema import to_openapi_json_schema
from .schema import validate_openapi_object

logger = getLogger(__name__)

MediaTypeKey = Literal['application/json', 'text/plain']


def get_media_type(body: str) -> Optional[MediaTypeKey]:
    """Determine media type (application/json, text/plain, etc.) from body.
    Return None for empty body.

    Arguments:
        body {str} -- Response body

    Raises:
        Exception: If body is of unexpected type.

    Returns:
        MediaTypeKey -- Media type such as "application/json"
    """

    if body == '':
        return None

    try:
        as_json = json.loads(body)
    except json.decoder.JSONDecodeError:
        logger.exception(f"Failed decoding: {body}")
        raise

    if isinstance(as_json, dict):
        return 'application/json'
    elif isinstance(as_json, list):
        return 'application/json'
    elif isinstance(as_json, str):
        return 'text/plain'
    else:
        raise Exception(f"Not sure what to do with body: {body}")


SchemaType = Literal['object', 'array', 'string']


def infer_schema(body: str, schema: Optional[Any] = None) -> Schema:
    try:
        as_json = json.loads(body)
    except json.decoder.JSONDecodeError:
        logger.exception(f"Failed decoding: {body}")
        raise

    # TODO typeguard
    return cast(Schema, to_openapi_json_schema(as_json, schema))


def update_media_type(request: HttpExchange, media_type: Optional[MediaType] = None) -> MediaType:
    body = request['res']['body']
    if media_type is not None:
        schema_or_none = media_type['schema']
    else:
        schema_or_none = None
    schema = infer_schema(body, schema=schema_or_none)
    media_type = MediaType(schema=schema)
    return media_type


def content_from_body(request: HttpExchange) -> Optional[Tuple[str, MediaType]]:
    body = request['res']['body']
    media_type_key = get_media_type(body)
    if media_type_key is None:
        return None
    media_type = update_media_type(request)
    return (media_type_key, media_type)


def build_response(request: HttpExchange) -> Response:
    """Build new response object from request response pair.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        request {HttpExchange} -- Request-response pair.

    Returns:
        Response -- OpenAPI response object.
    """
    # TODO Headers and links
    content_or_none = content_from_body(request)

    if content_or_none is None:
        return Response(
            description="Response description",
            headers={},
            links={}
        )

    media_type_key, media_type = content_or_none
    content = {
        media_type_key: media_type
    }

    return Response(
        description="Response description",
        headers={},
        content=content,
        links={}
    )


def update_response(response: Response, request: HttpExchange) -> Response:
    """Update response object. Mutates the input object.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        response {Response} -- Existing response object.
        request {HttpExchange} -- Request-response pair.

    Returns:
        Response -- Updated response object.
    """
    # TODO Update headers and links
    response_content = response['content'] if 'content' in response else None
    media_type_key = get_media_type(request['res']['body'])

    # No body
    if media_type_key is None:
        return response

    if response_content is not None and media_type_key in response_content:
        # Need to update media type
        existing_media_type = response_content[media_type_key]
        media_type = update_media_type(request, existing_media_type)
    else:
        media_type = update_media_type(request)

    response_content[media_type_key] = media_type
    return response


def build_operation(request: HttpExchange) -> Operation:
    """Build new operation object from request-response pair.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Operation object.
    """
    response = build_response(request)
    code = str(request['res']['statusCode'])
    operation = Operation(
        summary="Operation summary",
        description="Operation description",
        operationId="id",
        responses={code: response})
    return operation


def update_operation(operation: Operation, request: HttpExchange) -> Operation:
    """Update OpenAPI operation object. Mutates the input object.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        operation {Operation} -- Existing Operation object.
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Updated operation
    """
    responses = operation['responses']
    response_code = str(request['res']['statusCode'])
    if response_code in responses:
        # Response exists
        existing_response = responses[response_code]
        # Ensure response is Response and not Reference
        # check_type('existing_response', existing_response, Response)
        existing_response = cast(Response, existing_response)
        response = update_response(existing_response, request)
    else:
        response = build_response(request)

    operation['responses'][response_code] = response
    return operation


def update(schema: OpenAPIObject, request: HttpExchange) -> OpenAPIObject:
    """Update OpenAPI schema with a new request-response pair.
    Does not mutate the input schema.

    Schema reference: https://swagger.io/specification/#oasObject

    Returns:
        OpenAPI -- Updated schema
    """
    schema_copy = copy.deepcopy(schema)

    method = request['req']['method']

    if request['req']['path'] in schema_copy['paths']:
        # Path item exists for request path
        path_item: PathItem = schema_copy['paths'][request['req']['path']]
    else:
        path_item = PathItem(summary="Path summary",
                             description="Path description")
        schema_copy['paths'][request['req']['path']] = path_item

    if method in path_item:
        # Operation exists
        existing_operation = path_item[method]  # type: ignore
        operation = update_operation(existing_operation, request)
    else:
        operation = build_operation(request)

    # Needs type ignore as one cannot set variable property on typed dict
    path_item[method] = operation  # type: ignore

    return cast(OpenAPIObject, schema_copy)


BASE_SCHEMA = OpenAPIObject(openapi="3.0.0",
                            info=Info(title="API title",
                                      description="API description", version="1.0"),
                            paths={})


def build_schema_online(requests: Iterator[HttpExchange]) -> OpenAPIObject:
    """Build OpenAPI schema by iterating request-response pairs.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {Iterator[HttpExchange]} -- Iterator of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """

    # Iterate over all request-response pairs in the iterator, starting from
    # BASE_SCHEMA
    schema = reduce(lambda schema, req: update(
        schema, req), requests, BASE_SCHEMA)

    validate_openapi_object(schema)

    return schema


def build_schema_batch(requests: List[HttpExchange]) -> OpenAPIObject:
    """Build OpenAPI schema from a list of request-response object.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {List[HttpExchange]} -- List of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """
    return build_schema_online(iter(requests))



import copy
from meeshkan.schemabuilder import build_schema_batch
from meeshkan.schemabuilder.builder import BASE_SCHEMA
from meeshkan.schemabuilder.schema import validate_openapi_object
from ..util import read_recordings_as_request_response
from openapi_typed import OpenAPIObject, Operation, PathItem, Response, Schema
from typeguard import check_type
import pytest
from typing import cast
from hamcrest import *

requests = read_recordings_as_request_response()

expected_schema = copy.deepcopy(BASE_SCHEMA)

expected_schema['paths'] = {
    '/user/repos': PathItem(
        get=Operation(
            responses={
                '200': Response(description="description"),
                '403': Response(description="description")
            }
        )
    )
}


@pytest.fixture(scope="module")
def schema():
    return build_schema_batch(requests)


class TestSchema:
    schema = build_schema_batch(requests)

    def response_schema(self):

        return cast(Schema, self.schema['paths']['/user/repos']
                    ['get']['responses']['200']
                    ['content']['application/json']['schema'])  # type: ignore

    def test_schema_valid(self):
        validate_openapi_object(self.schema)

    def test_schema_typechecks(self):
        check_type('schema', self.schema, OpenAPIObject)

    def test_schema_keys(self):
        assert 'openapi' in self.schema
        assert 'info' in self.schema
        assert 'paths' in self.schema

    def test_paths(self):
        paths = self.schema['paths']
        assert '/user/repos' in paths

    def test_get_operation(self):
        op = self.schema['paths']['/user/repos']['get']
        assert 'responses' in op

    def test_get_operation_responses(self):
        responses = self.schema['paths']['/user/repos']['get']['responses']
        assert '200' in responses
        assert '403' in responses

    def test_200_response(self):
        response = self.schema['paths']['/user/repos']['get']['responses']['200']
        # TODO Typeguard
        response = cast(Response, response)
        assert response['headers'] == {}
        assert response['links'] == {}
        assert 'application/json' in response['content']
        media_type = response['content']['application/json']
        assert 'schema' in media_type

    def test_repos_content(self):
        content = cast(
            Response, self.schema['paths']['/user/repos']['get']['responses']['200'])['content']
        assert_that(content, has_key("application/json"))
        json_content = content['application/json']
        assert_that(json_content, has_key("schema"))
        schema = json_content['schema']
        assert_that(schema, has_entry("type", "array"))
        # TODO Typeguard for Reference
        schema = cast(Schema, schema)
        assert_that(schema, has_entry("items", instance_of(dict)))

    def test_items(self):
        schema = self.response_schema()
        items = schema['items']
        assert isinstance(items, dict)  # typeguard
        assert_that(items, has_entry('properties', instance_of(dict)))
        properties = items['properties']
        assert_that(properties, has_entry('clone_url',
                                          equal_to({'type': 'string'})))
